# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
from dataclasses import dataclass, field
from typing import List, Union

import geojson
from geojson import Feature
from pytdml.type.basic_types import Label, TrainingData, TrainingDataset, DataQuality, Task, \
    MetricsInLiterature, \
    KeyValuePair, Labeling, Changeset


@dataclass
class SceneLabel(Label):
    """
    Extended label type for scene level training data
    """
    label_class: str = field(default=None)

    def to_dict(self):
        return {
            "type": "SceneLabel",
            "isNegative": self.is_negative,
            "confidence": self.confidence,
            'class': self.label_class
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("class") and json_dict["type"] == "SceneLabel":
            label = SceneLabel(label_class=json_dict["class"])
            if json_dict.__contains__("isNegative"):
                label.is_negative = json_dict["isNegative"]
            if json_dict.__contains__("confidence"):
                label.confidence = json_dict["confidence"]
            return label
        else:
            raise ValueError("The given json_dict is not a valid SceneLabel")


@dataclass
class ObjectLabel(Label):
    """
    Extended label type for object level training data
    """
    object: Feature = field(default=None)
    label_class: str = field(default=None)
    bbox_type: str = field(default=None)
    is_difficultly_detectable: bool = field(default=None)
    date_time: str = field(default=None)

    def to_dict(self):
        return {
            "type": "ObjectLabel",
            "isNegative": self.is_negative,
            "confidence": self.confidence,
            'object': self.object,
            'bboxType': self.bbox_type,
            'class': self.label_class,
            'isDiffDetectable': self.is_difficultly_detectable,
            'dateTime': self.date_time
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("object") and json_dict["type"] == "ObjectLabel":
            label = ObjectLabel(object=geojson.loads(json_dict["object"].__str__().replace("'", "\"")))
            if json_dict.__contains__("isNegative"):
                label.is_negative = json_dict["isNegative"]
            if json_dict.__contains__("confidence"):
                label.confidence = json_dict["confidence"]
            if json_dict.__contains__("bboxType"):
                label.bbox_type = json_dict["bboxType"]
            if json_dict.__contains__("class"):
                label.label_class = json_dict["class"]
            if json_dict.__contains__("isDiffDetectable"):
                label.is_difficultly_detectable = json_dict["isDiffDetectable"]
            if json_dict.__contains__("dateTime"):
                label.date_time = json_dict["dateTime"]
            return label
        else:
            raise ValueError("The given json_dict is not a valid ObjectLabel")


@dataclass
class PixelLabel(Label):
    """
    Extended label type for pixel level training data
    """
    image_url: Union[str, List[str]] = field(default=None)
    image_format: Union[str, List[str]] = field(default=None)

    def to_dict(self):
        return {
            "type": "PixelLabel",
            "isNegative": self.is_negative,
            "confidence": self.confidence,
            "imageURL": self.image_url,
            "imageFormat": self.image_format
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("imageURL") and json_dict.__contains__("imageFormat") \
                and json_dict["type"] == "PixelLabel":
            label = PixelLabel(image_url=json_dict["imageURL"], image_format=json_dict["imageFormat"])
            if json_dict.__contains__("isNegative"):
                label.is_negative = json_dict["isNegative"]
            if json_dict.__contains__("confidence"):
                label.confidence = json_dict["confidence"]
            return label
        else:
            raise ValueError("The given json_dict is not a valid PixelLabel")


# @dataclass
# class EODataSource:
#     """
#     EO data source type
#     """
#     id: str
#     data_type: str = field(default=None)
#     citation: str = field(default=None)
#     platform: str = field(default=None)
#     sensor: str = field(default=None)
#     resolution: str = field(default=None)
#     format: str = field(default=None)
#
#     def to_dict(self):
#         return {
#             'id': self.id,
#             'dataType': self.data_type,
#             'citation': self.citation,
#             'platform': self.platform,
#             'sensor': self.sensor,
#             'resolution': self.resolution,
#             'format': self.format
#         }
#
#     @staticmethod
#     def from_dict(json_dict: dict):
#         if json_dict.__contains__("id"):
#             data_source = EODataSource(json_dict["id"])
#             if json_dict.__contains__("dataType"):
#                 data_source.data_type = json_dict["dataType"]
#             if json_dict.__contains__("citation"):
#                 data_source.citation = json_dict["citation"]
#             if json_dict.__contains__("platform"):
#                 data_source.platform = json_dict["platform"]
#             if json_dict.__contains__("sensor"):
#                 data_source.sensor = json_dict["sensor"]
#             if json_dict.__contains__("resolution"):
#                 data_source.resolution = json_dict["resolution"]
#             if json_dict.__contains__("format"):
#                     data_source.resolution = json_dict["format"]
#             return data_source
#         else:
#             raise ValueError("The given json_dict is not a valid EODataSource")


@dataclass
class EOTask(Task):
    """
    Extended task type for EO training data
    """
    task_type: str = field(default=None)

    def to_dict(self):
        return {
            "type": "EOTask",
            "id": self.id,
            "datasetId": self.dataset_id,
            "description": self.description,
            "taskType": self.task_type
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict.__contains__("type") and json_dict["type"] == "EOTask":
            task = EOTask(json_dict["id"])
            if json_dict.__contains__("datasetId"):
                task.dataset_id = json_dict["datasetId"]
            if json_dict.__contains__("description"):
                task.description = json_dict["description"]
            if json_dict.__contains__("taskType"):
                task.task_type = json_dict["taskType"]
            return task
        else:
            raise ValueError("The given json_dict is not a valid EOTask")


@dataclass
class EOTrainingData(TrainingData):
    """
    Extended training data type for EO training data
    """
    extent: List[float] = field(default_factory=list)
    date_time: Union[str, List[str]] = field(default=None)
    data_url: Union[str, List[str]] = field(default=None)

    def to_dict(self):
        return {
            "type": "EOTrainingData",
            "id": self.id,
            "trainingType": self.training_type,
            "numberOfLabels": self.number_of_labels,
            "dataSources": self.data_sources,
            'extent': self.extent,
            'dateTime': self.date_time,
            'dataURL': self.data_url,
            "quality": self.quality.to_dict() if self.quality is not None else None,
            "labeling": [ll.to_dict() for ll in self.labeling],
            "labels": [label.to_dict() for label in self.labels]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict["type"] == "EOTrainingData":
            training_data = EOTrainingData(json_dict["id"])
            if json_dict.__contains__("trainingType"):
                training_data.training_type = json_dict["trainingType"]
            if json_dict.__contains__("numberOfLabels"):
                training_data.number_of_labels = json_dict["numberOfLabels"]
            if json_dict.__contains__("dataSources"):
                training_data.data_sources = json_dict["dataSources"]
            if json_dict.__contains__("quality"):
                training_data.quality = DataQuality.from_dict(json_dict["quality"])
            if json_dict.__contains__("labeling"):
                training_data.labeling = [Labeling.from_dict(ll) for ll in json_dict["labeling"]]
            if json_dict.__contains__("labels"):
                # Different type of labels
                if json_dict["labels"][0]["type"] == "SceneLabel":
                    training_data.labels = [SceneLabel.from_dict(label) for label in json_dict["labels"]]
                elif json_dict["labels"][0]["type"] == "ObjectLabel":
                    training_data.labels = [ObjectLabel.from_dict(label) for label in json_dict["labels"]]
                elif json_dict["labels"][0]["type"] == "PixelLabel":
                    training_data.labels = [PixelLabel.from_dict(label) for label in json_dict["labels"]]
            if json_dict.__contains__("extent"):
                training_data.extent = json_dict["extent"]
            if json_dict.__contains__("dateTime"):
                training_data.date_time = json_dict["dateTime"]
            if json_dict.__contains__("dataURL"):
                training_data.data_url = json_dict["dataURL"]
            return training_data
        else:
            raise ValueError("The given json_dict is not a valid EOTrainingData")


@dataclass
class EOTrainingDataset(TrainingDataset):
    """
    Extended training dataset type for EO training dataset
    """
    extent: List[float] = field(default_factory=list)
    bands: List[str] = field(default_factory=list)
    image_size: str = field(default=None)

    def to_dict(self):
        return {
            "type": "EOTrainingDataset",
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "license": self.license,
            "doi": self.doi,
            "scope": self.scope.to_dict() if self.scope is not None else None,
            "version": self.version,
            "amountOfTrainingData": self.amount_of_training_data,
            "createdTime": self.created_time,
            "updatedTime": self.updated_time,
            "providers": self.providers,
            "keywords": self.keywords,
            "metricsInLIT": [m.to_dict() for m in self.metrics_in_literature],
            "statisticsInfo": [s.to_dict() for s in self.statistics_info],
            "numberOfClasses": self.number_of_classes,
            "classificationSchema": self.classification_schema,
            "classes": self.classes,
            "tasks": [t.to_dict() for t in self.tasks],
            'extent': self.extent,
            'bands': self.bands,
            'imageSize': self.image_size,
            "labeling": [labeling.to_dict() for labeling in self.labeling],
            "quality": self.quality.to_dict() if self.quality is not None else None,
            "changesets": [changeset.to_dict() for changeset in self.changesets],
            "data": [d.to_dict() for d in self.data]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict.__contains__("name") \
                and json_dict.__contains__("description") and json_dict.__contains__("license") \
                and json_dict["type"] == "EOTrainingDataset":
            td = EOTrainingDataset(json_dict["id"], json_dict["name"], json_dict["description"], json_dict["license"])
            if json_dict.__contains__("doi"):
                td.doi = json_dict["doi"]
            if json_dict.__contains__("scope"):
                td.scope = Scope.from_dict(json_dict["scope"])
            if json_dict.__contains__("version"):
                td.version = json_dict["version"]
            if json_dict.__contains__("amountOfTrainingData"):
                td.amount_of_training_data = json_dict["amountOfTrainingData"]
            if json_dict.__contains__("createdTime"):
                td.created_time = json_dict["createdTime"]
            if json_dict.__contains__("updatedTime"):
                td.updated_time = json_dict["updatedTime"]
            if json_dict.__contains__("providers"):
                td.providers = json_dict["providers"]
            if json_dict.__contains__("keywords"):
                td.keywords = json_dict["keywords"]
            if json_dict.__contains__("metricsInLIT"):
                td.metrics_in_literature = [MetricsInLiterature.from_dict(m) for m in
                                            json_dict["metricsInLIT"]]
            if json_dict.__contains__("statisticsInfo"):
                td.statistics_info = [KeyValuePair.from_dict(s) for s in json_dict["statisticsInfo"]]
            if json_dict.__contains__("numberOfClasses"):
                td.number_of_classes = json_dict["numberOfClasses"]
            if json_dict.__contains__("classificationSchema"):
                td.classification_schema = json_dict["classificationSchema"]
            if json_dict.__contains__("classes"):
                td.classes = json_dict["classes"]
            if json_dict.__contains__("tasks"):
                td.tasks = [EOTask.from_dict(t) for t in json_dict["tasks"]]
            if json_dict.__contains__("labeling"):
                td.labeling = [Labeling.from_dict(labeling) for labeling in json_dict["labeling"]]
            if json_dict.__contains__("quality"):
                td.quality = DataQuality.from_dict(json_dict["quality"])
            if json_dict.__contains__("changesets"):
                td.changesets = [Changeset.from_dict(changeset) for changeset in json_dict["changesets"]]
            if json_dict.__contains__("data"):
                td.data = [EOTrainingData.from_dict(data) for data in json_dict["data"]]
            if json_dict.__contains__("extent"):
                td.extent = json_dict["extent"]
            if json_dict.__contains__("bands"):
                td.bands = json_dict["bands"]
            if json_dict.__contains__("imageSize"):
                td.image_size = json_dict["imageSize"]
            return td
        else:
            raise ValueError("The given json_dict is not a valid EOTrainingDataset")



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import random

from pytdml.type import TrainingDataset


def split_train_valid_test(td: TrainingDataset, train_ratio=0.7, valid_ratio=0.0, test_ratio=0.3):
    """
    Splits a TrainingDataset into train, valid and test sets.
    """
    assert abs(train_ratio + valid_ratio + test_ratio - 1.0) <= 0.000001, "Ratios must sum to 1.0"

    n_samples = len(td.data)
    n_train = int(n_samples * train_ratio)
    n_valid = int(n_samples * valid_ratio)

    random.shuffle(td.data)

    train_td = td.data[:n_train]
    valid_td = td.data[n_train:n_train + n_valid]
    test_td = td.data[n_train + n_valid:]

    return train_td, valid_td, test_td


def creat_class_map(td: TrainingDataset):
    """
    Creates a map from class labels to indices.
    """
    assert td.classes is not None, "Training dataset has no classes"
    class_map = {}
    for i, _class in enumerate(td.classes):
        if isinstance(_class, dict):
            for item in _class.items():
                class_map[item[0]] = item[1]
        else:
            class_map[_class] = i
    return class_map



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import argparse
import json
import os
import sys

import cv2

from pytdml.io import read_from_json
from pytdml.type import EOTrainingDataset, EOTrainingData, PixelLabel
from pytdml.utils import remove_empty


def td_image_crop(td: EOTrainingDataset, save_tdml_path: str, save_crop_dir: str, sub_size: int):
    td_dict = td.to_dict()
    td_list = td_dict['data']
    new_td_list = []
    index = 0
    for d in td_list:
        label_type = d['labels'][0]['type']
        if label_type == 'PixelLabel':
            image_url = d['dataURL']
            label_url = d['labels'][0]['imageURL']
            image_dir = os.path.join(save_crop_dir, "images")
            label_dir = os.path.join(save_crop_dir, "labels")
            if not os.path.isdir(image_dir):
                os.makedirs(image_dir)
            if not os.path.isdir(label_dir):
                os.makedirs(label_dir)
            crop_image_list = image_crop(image_url, image_dir, sub_size)
            crop_label_list = image_crop(label_url, label_dir, sub_size)
            for crop_image_url, crop_label_url in zip(crop_image_list, crop_label_list):
                new_d = EOTrainingData(
                    id=str(index),
                    labels=[PixelLabel(image_url=crop_label_url)],
                    data_url=crop_image_url,
                )
                index = index + 1
                new_td_list.append(new_d.to_dict())
    td_dict["amountOfTrainingData"] = len(new_td_list)
    td_dict['data'] = new_td_list
    with open(save_tdml_path, "w") as f:
        json.dump(remove_empty(td_dict), f, indent=4)


def image_crop(data_url, save_dir, sub_size):
    """
    image crop form data url
    """
    image = cv2.imread(data_url)
    image_name = os.path.basename(data_url)
    h = image.shape[0]
    w = image.shape[1]
    crop_image_path_list = []
    if not os.path.isdir(save_dir):
        os.makedirs(save_dir)
    if h > sub_size & (h > sub_size):
        for i in range(int(w / sub_size)):
            for j in range(int(h / sub_size)):
                crop_image_path = os.path.join(save_dir, str(i) + str(j) + image_name)
                crop_image = image[j * sub_size: sub_size * (j + 1), i * sub_size: sub_size * (i + 1)]
                cv2.imwrite(crop_image_path, crop_image)
                crop_image_path_list.append(crop_image_path)
    return crop_image_path_list


def main():
    parser = argparse.ArgumentParser(description='Encode a training dataset to TrainingDML-AI JSON format based on '
                                                 'YAML configuration file')
    parser.add_argument('--input', type=str, required=True, help='Input original TrainingDML-AU file path')
    parser.add_argument('--output_json', type=str, required=True, help='Output result TrainingDML-AI JSON file path')
    parser.add_argument('--output_images', type=str, required=True, help='Output dir of result cropped images')
    parser.add_argument('--size', type=int, required=True, help='Crop size of images')

    args = parser.parse_args()
    tdml_path = args.input
    save_tdml_path = args.output_json
    save_crop_dir = args.output_images
    sub_size = args.size
    training_datasets = read_from_json(tdml_path)
    if training_datasets:
        td_image_crop(training_datasets, save_tdml_path, save_crop_dir, sub_size)


if __name__ == '__main__':
    sys.exit(main())



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import argparse
import json
import os
import sys

import yaml
import pandas as pd
from geojson import Polygon, Feature

from pytdml.io import write_to_json
from pytdml.type import EOTask, EOTrainingDataset, EOTrainingData, SceneLabel, PixelLabel, ObjectLabel, Scope


def yaml_to_eo_tdml(yaml_path):
    """
    Transform yaml to tdml
    """
    yaml_file = open(yaml_path, "r", encoding='utf-8')
    yaml_dict = yaml.load(yaml_file, Loader=yaml.FullLoader)
    dataset_type = yaml_dict['dataset_type']
    tasks_list = yaml_dict['tasks']
    td_list = load_data(yaml_dict['data'])

    tasks = []
    for task in tasks_list:
        eo_task = EOTask(
            id=task.__contains__('id') and task['id'] or "",
            dataset_id=task.__contains__('dataset_id') and task['dataset_id'] or "",
            task_type=task.__contains__('task_type') and task['task_type'] or "",
            description=task.__contains__('description') and task['description'] or ""
        )
        tasks.append(eo_task)

    if dataset_type == 'EOTrainingDataset':
        eo_training_dataset = EOTrainingDataset(
            id=yaml_dict['id'],
            name=yaml_dict['name'],
            description=yaml_dict['description'],
            license=yaml_dict['license'],
            doi=yaml_dict['doi'],
            scope=Scope.from_dict(yaml_dict['scope']),
            version=yaml_dict['version'],
            amount_of_training_data=len(td_list),
            created_time=yaml_dict['created_time'],
            updated_time=yaml_dict['updated_time'],
            providers=yaml_dict["providers"],
            keywords=yaml_dict["keywords"],
            classification_schema=yaml_dict['classification_schema'],
            classes=yaml_dict['classes'],
            tasks=tasks,
            extent=yaml_dict['extent'],
            bands=yaml_dict['bands'],
            image_size=yaml_dict['image_size'],
            data=td_list,
        )
        return eo_training_dataset


def load_data(data_dict):
    """
    Load training dataset
    """
    task_type = data_dict['task_type']
    label_type = data_dict['label_type']
    data_path = data_dict['data_path']

    if label_type == 'SceneLabel' and task_type == 'SceneClassification':
        for data in data_path:
            if data['type'] == 'image':
                return load_data_scene_label(data)
    elif label_type == 'ObjectLabel' and task_type == 'ObjectDetection':
        image_set = {}
        label_set = {}
        for data in data_path:
            if data['type'] == 'image':
                image_set = data
            elif data['type'] == 'label':
                label_set = data
        return load_data_object_detection(image_set, label_set)
    elif label_type == 'PixelLabel' and task_type == 'SemanticSegmentation':
        image_set = {}
        label_set = {}
        for data in data_path:
            if data['type'] == 'image':
                image_set = data
            elif data['type'] == 'label':
                label_set = data
        return load_data_semantic_segmentation(image_set, label_set)
    elif label_type == 'PixelLabel' and task_type == 'ChangeDetection':
        image_before_set = {}
        image_after_set = {}
        label_set = {}
        data_time = data_dict['data_time']
        for data in data_path:
            if data['type'] == 'image_before':
                image_before_set = data
            elif data['type'] == 'image_after':
                image_after_set = data
            elif data['type'] == 'label':
                label_set = data
        return load_data_change_detection(image_before_set, image_after_set, label_set, data_time)


def load_data_scene_label(image_set):
    """
    Load training dataset for scene classification
    """
    td_list = []
    file_format = image_set['format']
    root_path = image_set['root_path']
    sub_dirs = []
    if image_set.__contains__('sub_path') and image_set['sub_path'] is not None:  # if set sub path
        sub_dirs = image_set['sub_path']
    else:
        for root, dirs, files in os.walk(root_path):
            sub_dirs = dirs
            break
    try:
        for sub_dir in sub_dirs:
            image_path = os.path.join(root_path, sub_dir)
            for root, dirs, files in os.walk(image_path):
                for file in files:
                    root = str(root)
                    file = str(file)
                    data_url = os.path.join(root, file)
                    if file_format == os.path.splitext(data_url)[-1]:
                        td = EOTrainingData(
                            id=file.split(".")[0],
                            labels=[SceneLabel(label_class=sub_dir)],
                            data_url=data_url,
                        )
                        td_list.append(td)
        return td_list
    except IOError:
        return IOError("Failed to read scene label")


def load_data_semantic_segmentation(image_set, label_set):
    """
    Load training dataset for semantic segmentation
    """
    td_list = []
    image_dir, label_dir = get_image_label_url(image_set, label_set)
    index = 0
    for image_url, label_url in zip(image_dir, label_dir):
        td = EOTrainingData(
            id=str(index),
            labels=[PixelLabel(image_url=label_url)],
            data_url=str(image_url)
        )
        index = index + 1
        td_list.append(td)
    return td_list


def load_data_change_detection(image_before_set, image_after_set, label_set, data_time):
    """
    Load training dataset for change detection
    """
    td_list = []
    image_before_root_path = image_before_set['root_path']
    image_after_root_path = image_after_set['root_path']
    label_root_path = label_set['root_path']
    image_before_format = image_before_set['format']
    image_after_format = image_after_set['format']
    label_format = label_set['format']
    image_before_dir = []
    image_after_dir = []
    label_dir = []

    if image_before_set.__contains__('sub_path') and image_before_set['sub_path'] is not None \
            and image_after_set.__contains__('sub_path') and image_after_set['sub_path'] is not None \
            and label_set.__contains__('sub_path') and label_set['sub_path'] is not None:
        image_before_sub_paths = image_before_set['sub_path']
        image_after_sub_paths = image_after_set['sub_path']
        label_sub_paths = label_set['sub_path']
        for image_before_sub_path in image_before_sub_paths:
            image_before_dir.extend(
                traverse_folder(os.path.join(image_before_root_path, image_before_sub_path), image_before_format))
        for image_after_sub_path in image_after_sub_paths:
            image_after_dir.extend(
                traverse_folder(os.path.join(image_after_root_path, image_after_sub_path), image_after_format))
        for label_sub_path in label_sub_paths:
            label_dir.extend(traverse_folder(os.path.join(label_root_path, label_sub_path), label_format))
    else:
        image_before_dir = traverse_folder(image_before_root_path, image_before_format)
        image_after_dir = traverse_folder(image_after_root_path, image_after_format)
        label_dir = traverse_folder(label_root_path, label_format)

    index = 0
    for image_before_url, image_after_url, label_url in zip(image_before_dir, image_after_dir, label_dir):
        td = EOTrainingData(
            id=str(index),
            labels=[PixelLabel(image_url=label_url)],
            data_url=[image_before_url, image_after_url]
        )
        td.date_time = [str(data_time[0]), str(data_time[1])]
        index = index + 1
        td_list.append(td)
    return td_list


def load_data_object_detection(image_set, label_set):
    """
    Load training dataset for object detection
    """
    td_list = []
    label_format = label_set['format']
    image_format = image_set['format']

    if label_format == '.txt':  # txt format
        image_dir, label_dir = get_image_label_url(image_set, label_set)
        separate = label_set['separate']
        label_column = label_set['column']
        index = 0
        for image_url, label_url in zip(image_dir, label_dir):
            labels = read_txt_label(label_url, label_column, separate)
            td = EOTrainingData(
                id=str(index),
                labels=labels,
                data_url=image_url,
            )
            td.number_of_labels = len(td.labels)
            index = index + 1
            td_list.append(td)
    elif label_format == 'stac' and image_format == 'stac':  # stac format
        images_stac_path = image_set['root_path'] + "\\collection.json"
        labels_stac_path = label_set['root_path'] + "\\collection.json"
        images_stac_json = json.load(open(images_stac_path, 'r'))
        labels_stac_json = json.load(open(labels_stac_path, 'r'))
        num = len(images_stac_json['links'])
        for i in range(num):
            image_stac_path = os.path.join(image_set['root_path'], images_stac_json['links'][i]['href'])
            label_stac_path = os.path.join(label_set['root_path'], labels_stac_json['links'][i]['href'])
            image_stac_json = json.load(open(image_stac_path, 'r'))
            label_stac_json = json.load(open(label_stac_path, 'r'))
            label_json_path = os.path.join(os.path.dirname(label_stac_path),
                                           label_stac_json['assets']['labels']['href'])
            labels = read_geojson_label(label_json_path)
            td = EOTrainingData(
                id=image_stac_json['id'],
                extent=image_stac_json['bbox'],
                date_time=image_stac_json['properties']['datetime'],
                data_url=os.path.join(os.path.dirname(image_stac_path),
                                      list(image_stac_json['assets'].values())[0]['href']),
                labels=labels,
            )
            td.number_of_labels = len(td.labels)
            td_list.append(td)
        return td_list


def read_txt_label(csv_path, column_name, separate):
    """
    Read the label file in text format in object detection
    """
    df = pd.read_csv(csv_path, sep=separate, header=None, names=column_name, engine='python')
    labels = []
    is_have_nan = df.isnull().any(axis=1)  # if null
    for index, row in df.iterrows():
        if not is_have_nan[index]:
            polygon = Polygon(
                [(float(row['x1']), float(row['y1'])),
                 (float(row['x2']), float(row['y2'])),
                 (float(row['x3']), float(row['y3'])),
                 (float(row['x4']), float(row['y4']))]
            )
            label = ObjectLabel(
                object=Feature(geometry=polygon),
                bbox_type='Horizontal BBox',
                label_class=row['class'],
                is_difficultly_detectable=row['isDiffDetectable']
            )
            labels.append(label)
    return labels


def read_geojson_label(geojson_path):
    """
    Read the label file in geojson format in object detection
    """
    fc = json.load(open(geojson_path))
    labels = []
    for f in fc['features']:
        labels.append(ObjectLabel(
            object=f
        ))
    return labels


def traverse_folder(file_path, file_format):
    """
    Traverse target folder
    """
    try:
        file_dir = []
        for root, dirs, files in os.walk(file_path):
            for f in files:
                if file_format == os.path.splitext(f)[-1]:
                    file_dir.append(os.path.join(root, f))
        return file_dir
    except IOError:
        return IOError("Failed to load dataset")


def get_image_label_url(image_set, label_set):
    """
    Get image url and label url from image set and label set
    """
    image_root_path = image_set['root_path']
    label_root_path = label_set['root_path']
    image_format = image_set['format']
    label_format = label_set['format']
    image_dir = []
    label_dir = []
    # if set sub path
    if image_set.__contains__('sub_path') and image_set['sub_path'] is not None \
            and label_set.__contains__('sub_path') and label_set['sub_path'] is not None:
        image_sub_paths = image_set['sub_path']
        label_sub_paths = label_set['sub_path']
        for image_sub_path in image_sub_paths:
            image_dir.extend(traverse_folder(os.path.join(image_root_path, image_sub_path), image_format))
        for label_sub_path in label_sub_paths:
            label_dir.extend(traverse_folder(os.path.join(label_root_path, label_sub_path), label_format))
    else:
        image_dir = traverse_folder(image_root_path, image_format)
        label_dir = traverse_folder(label_root_path, label_format)
    return image_dir, label_dir


def main():
    parser = argparse.ArgumentParser(description='Encode a training dataset to TrainingDML-AI JSON format based on '
                                                 'YAML configuration file')
    parser.add_argument('--config', type=str, required=True, help='YAML configuration file path')
    parser.add_argument('--output', type=str, required=True, help='Output TrainingDML-AI JSON file path')

    args = parser.parse_args()
    yaml_path = args.config
    json_path = args.output
    training_datasets = yaml_to_eo_tdml(yaml_path)
    if training_datasets:
        write_to_json(training_datasets, json_path)


if __name__ == '__main__':
    sys.exit(main())



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import cv2
import numpy as np
import torch
from PIL import Image

import pytdml.utils as utils

from torch.utils.data import Dataset


class TorchEOImageSceneTD(Dataset):
    """
    Torch Dataset for EO image scene classification training dataset
    """

    def __init__(self, td_list, class_map, transform=None):
        self.td_list = td_list
        self.class_map = class_map
        self.transform = transform

    def __len__(self):
        return len(self.td_list)

    def __getitem__(self, item):
        img_path = self.td_list[item].data_url
        img = Image.open(img_path)
        img.load()
        label = self.class_map[self.td_list[item].labels[0].label_class]
        if self.transform is not None:
            img = self.transform(img)
        return img, label


class TorchEOImageObjectTD(Dataset):
    """
    Torch Dataset for EO image object detection training dataset
    """

    def __init__(self, td_list, class_map, transform=None):
        self.td_list = td_list
        self.class_map = class_map
        self.transform = transform

    def __len__(self):
        return len(self.td_list)

    def __getitem__(self, index):
        img_path = self.td_list[index].data_url
        img = cv2.imread(img_path)
        img_height, img_width, channels = img.shape
        targets = []
        labels = self.td_list[index].labels
        # transform annotations
        for label in labels:
            class_value_ = self.class_map[label.label_class]
            target_ = utils.get_bounding_box(label.object)
            target_[0] = target_[0] / img_width
            target_[1] = target_[1] / img_height
            target_[2] = target_[2] / img_width
            target_[3] = target_[3] / img_height
            target_.append(class_value_)
            targets.append(target_)
        targets = np.array(targets)
        img = img[:, :, (2, 1, 0)]
        if self.transform is not None:
            img, boxes, labels = self.transform(img, targets[:, :4], targets[:, 4])
            targets = np.hstack((boxes, np.expand_dims(labels, axis=1)))
        return torch.from_numpy(img).permute(2, 0, 1), targets, img_height, img_width


class TorchEOImageSegmentationTD(Dataset):
    """
    Torch Dataset for EO image Semantic Segmentation training dataset
    """

    def __init__(self, td_list, class_map, transform):
        self.td_list = td_list
        self.class_map = class_map
        self.transform = transform
        self.color_to_index = utils.class_to_index(class_map)

    def __len__(self):
        return len(self.td_list)

    def __getitem__(self, item):
        img_path = self.td_list[item].data_url
        img = Image.open(img_path)
        img.load()
        if self.transform is not None:
            img = self.transform(img)
        label_path = self.td_list[item].labels[0].image_url
        label = cv2.imread(label_path, cv2.COLOR_BGR2RGB)
        index_label = utils.label_to_index(label, self.color_to_index)
        index_label = np.asarray(index_label, dtype=np.int64)
        index_label = torch.from_numpy(index_label)
        return img, index_label


def base_transform(image, size, mean, std):
    x = cv2.resize(image, (size[0], size[1]), interpolation=cv2.INTER_AREA).astype(np.float32)
    x /= 255.
    x -= mean
    x /= std
    return x


class BaseTransform:
    def __init__(self, size=None, mean=(0.406, 0.456, 0.485), std=(0.225, 0.224, 0.229)):
        if size is None:
            size = [256, 256]
        self.size = size
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)

    def __call__(self, image, boxes=None, labels=None):
        return base_transform(image, self.size, self.mean, self.std), boxes, labels



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
from dataclasses import dataclass, field
from typing import List, Union


@dataclass
class KeyValuePair:
    """
    Key/Value pair type
    """
    key: str
    value: Union[str, int, float, bool, None]

    def to_dict(self):
        return {self.key: self.value}

    @staticmethod
    def from_dict(json_dict: dict):
        if len(json_dict.keys()) == 1:
            for key in json_dict.keys():
                return KeyValuePair(key, json_dict[key])
        else:
            raise ValueError("The given json_dict is not a valid KeyValuePair")


@dataclass
class MetricsInLiterature:
    """
    Metrics in literature type
    """
    doi: str
    algorithm: str = field(default=None)
    metrics: List[KeyValuePair] = field(default_factory=list[KeyValuePair])

    def to_dict(self):
        return {
            "doi": self.doi,
            "algorithm": self.algorithm,
            "metrics": [m.to_dict() for m in self.metrics]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("doi"):
            metrics_in_literature = MetricsInLiterature(json_dict["doi"])
            if json_dict.__contains__("algorithm"):
                metrics_in_literature.algorithm = json_dict["algorithm"]
            if json_dict.__contains__("metrics"):
                metrics_in_literature.metrics = [KeyValuePair.from_dict(m) for m in json_dict["metrics"]]
            return metrics_in_literature
        else:
            raise ValueError("The given json_dict is not a valid MetricsInLiterature")


@dataclass
class Task:
    """
    Basic task type
    """
    id: str
    dataset_id: str = field(default=None)
    description: str = field(default=None)

    def to_dict(self):
        return {
            "type": "Task",
            "id": self.id,
            "datasetId": self.dataset_id,
            "description": self.description
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict["type"] == "Task":
            task = Task(json_dict["id"])
            if json_dict.__contains__("datasetId"):
                task.dataset_id = json_dict["datasetId"]
            if json_dict.__contains__("description"):
                task.description = json_dict["description"]
            return task
        else:
            raise ValueError("The given json_dict is not a valid Task")


@dataclass
class Labeler:
    """
    Labeler type
    """
    id: str
    name: str = field(default=None)

    def to_dict(self):
        return {
            "type": "Labeler",
            "id": self.id,
            "name": self.name
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict["type"] == "Labeler":
            labeler = Labeler(json_dict["id"])
            if json_dict.__contains__("name"):
                labeler.name = json_dict["name"]
            return labeler
        else:
            raise ValueError("The given json_dict is not a valid Labeler")


@dataclass
class LabelingProcedure:
    """
    Labeling procedure type
    """
    id: str
    methods: list[str] = field(default_factory=list)
    tools: list[str] = field(default_factory=list)

    def to_dict(self):
        return {
            "type": "LabelingProcedure",
            "id": self.id,
            "methods": self.methods,
            "tools": self.tools
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict["type"] == "LabelingProcedure":
            labeling_procedure = LabelingProcedure(json_dict["id"])
            if json_dict.__contains__("methods"):
                labeling_procedure.method = json_dict["methods"]
            if json_dict.__contains__("tools"):
                labeling_procedure.tool = json_dict["tools"]
            return labeling_procedure
        else:
            raise ValueError("The given json_dict is not a valid LabelingProcedure")


@dataclass
class Labeling:
    """
    Labeling type
    """
    id: str
    scope: object
    labelers: List[Labeler] = field(default_factory=list)
    procedure: LabelingProcedure = field(default=None)

    def to_dict(self):
        return {
            "type": "Labeling",
            "id": self.id,
            "scope": self.scope,
            "labelers": [labeler.to_dict() for labeler in self.labelers],
            "procedure": self.procedure.to_dict() if self.procedure is not None else None,
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict.__contains__("scope") and json_dict["type"] == "Labeling":
            labeling = Labeling(json_dict["id"], json_dict["scope"])
            if json_dict.__contains__("labelers"):
                labeling.labelers = [Labeler.from_dict(labeler) for labeler in json_dict["labelers"]]
            if json_dict.__contains__("procedure"):
                labeling.procedure = LabelingProcedure.from_dict(json_dict["procedure"])
            return labeling
        else:
            raise ValueError("The given json_dict is not a valid Labeling")


@dataclass
class ScopeDescription:
    """
    From ISO 19115-1 MD_ScopeDescription
    """
    attributes: list[str] = field(default_factory=list)
    features: list[str] = field(default_factory=list)
    featureInstances: list[str] = field(default_factory=list)
    attributeInstances: list[str] = field(default_factory=list)
    dataset: str = field(default=None)
    other: str = field(default=None)

    def to_dict(self):
        return {
            "attributes": self.attributes,
            "features": self.features,
            "featureInstances": self.featureInstances,
            "attributeInstances": self.attributeInstances,
            "dataset": self.dataset,
            "other": self.other
        }

    @staticmethod
    def from_dict(json_dict: dict):
        sd = ScopeDescription()
        if json_dict.__contains__("attributes"):
            sd.attributes = json_dict["attributes"]
        if json_dict.__contains__("features"):
            sd.features = json_dict["features"]
        if json_dict.__contains__("featureInstances"):
            sd.featureInstances = json_dict["featureInstances"]
        if json_dict.__contains__("attributeInstances"):
            sd.attributeInstances = json_dict["attributeInstances"]
        if json_dict.__contains__("dataset"):
            sd.dataset = json_dict["dataset"]
        if json_dict.__contains__("other"):
            sd.other = json_dict["other"]
        return sd


@dataclass
class Scope:
    """
    From ISO 19115-1 MD_Scope
    """
    level: str
    levelDescription: List[ScopeDescription] = field(default_factory=list[ScopeDescription])

    def to_dict(self):
        return {
            "level": self.level,
            "levelDescription": [ld.to_dict for ld in self.levelDescription]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("level"):
            scope = Scope(json_dict["level"])
            if json_dict.__contains__("levelDescription"):
                scope.levelDescription = [ScopeDescription.from_dict(ld) for ld in json_dict["levelDescription"]]
            return scope
        else:
            raise ValueError("The given json_dict is not a valid Scope")


@dataclass
class QualityElement:
    """
    From ISO 19157-1 QualityElement
    """
    type: str = field(default=None)
    measure: str = field(default=None)
    evaluation_method: str = field(default=None)
    result: str = field(default=None)

    def to_dict(self):
        return {
            "type": self.type,
            "measure": self.measure,
            "evaluationMethod": self.evaluation_method,
            "result": self.result
        }

    @staticmethod
    def from_dict(json_dict: dict):
        quality_element = QualityElement()
        if json_dict.__contains__("type"):
            quality_element.type = json_dict["type"]
        if json_dict.__contains__("measure"):
            quality_element.measure = json_dict["measure"]
        if json_dict.__contains__("evaluationMethod"):
            quality_element.evaluation_method = json_dict["evaluationMethod"]
        if json_dict.__contains__("result"):
            quality_element.result = json_dict["resule"]
        return quality_element


@dataclass
class DataQuality:
    """
    From ISO 19157-1 DataQuality
    """
    scope: Scope
    report: List[QualityElement] = field(default_factory=list[QualityElement])

    def to_dict(self):
        return {
            "scope": self.scope.to_dict() if self.scope is not None else None,
            "report": [r.to_dict() for r in self.report],
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("scope"):
            data_quality = DataQuality(scope=Scope.from_dict(json_dict["scope"]))
            if json_dict.__contains__("report"):
                data_quality.report = [QualityElement.from_dict(element) for element in json_dict["report"]]
            return data_quality
        else:
            raise ValueError("The given json_dict is not a valid DataQuality")


@dataclass
class Label:
    """
    Basic label type
    """
    is_negative: bool = field(default=None)
    confidence: float = field(default=None)

    def to_dict(self):
        return {
            "type": "Label",
            "isNegative": self.is_negative,
            "confidence": self.confidence
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict["type"] == "Label":
            label = Label()
            if json_dict.__contains__("isNegative"):
                label.is_negative = json_dict["isNegative"]
            if json_dict.__contains__("confidence"):
                label.confidence = json_dict["confidence"]
            return label
        else:
            raise ValueError("The given json_dict is not a valid Label")


@dataclass
class TrainingData:
    """
    Basic training data type
    """
    id: str
    dataset_id: str = field(default=None)
    training_type: str = field(default=None)
    number_of_labels: int = field(default=None)
    data_sources: List[str] = field(default_factory=list)
    quality: DataQuality = field(default=None)
    labeling: List[Labeling] = field(default_factory=list[Labeling])
    labels: List[Label] = field(default_factory=list[Label])

    def get_labels(self) -> List[Label]:
        """
        Returns the labels of the training data
        """
        return self.labels

    def to_dict(self):
        return {
            "type": "TrainingData",
            "id": self.id,
            "trainingType": self.training_type,
            "numberOfLabels": self.number_of_labels,
            "dataSources": self.data_sources,
            "quality": self.quality.to_dict() if self.quality is not None else None,
            "labeling": [ll.to_dict() for ll in self.labeling],
            "labels": [label.to_dict() for label in self.labels]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict["type"] == "TrainingData":
            training_data = TrainingData(json_dict["id"])
            if json_dict.__contains__("trainingType"):
                training_data.training_type = json_dict["trainingType"]
            if json_dict.__contains__("numberOfLabels"):
                training_data.number_of_labels = json_dict["numberOfLabels"]
            if json_dict.__contains__("dataSources"):
                training_data.data_sources = json_dict["dataSources"]
            if json_dict.__contains__("quality"):
                training_data.quality = DataQuality.from_dict(json_dict["quality"])
            if json_dict.__contains__("labeling"):
                training_data.labeling = [Labeling.from_dict(ll) for ll in json_dict["labeling"]]
            if json_dict.__contains__("labels"):
                training_data.labels = [Label.from_dict(label) for label in json_dict["labels"]]
            return training_data
        else:
            raise ValueError("The given json_dict is not a valid TrainingData")


@dataclass
class Changeset:
    """
    Training Data Changeset
    """
    id: str
    change_count: int
    dataset_id: str = field(default=None)
    version: str = field(default=None)
    created_time: str = field(default=None)
    add: List[TrainingData] = field(default_factory=list[TrainingData])
    modify: List[TrainingData] = field(default_factory=list[TrainingData])
    delete: List[TrainingData] = field(default_factory=list[TrainingData])

    def to_dict(self):
        return {
            "type": "TDChangeset",
            "id": self.id,
            "changeCount": self.change_count,
            "datasetId": self.dataset_id,
            "version": self.version,
            "createdTime": self.created_time,
            "add": [td.to_dict() for td in self.add],
            "modify": [td.to_dict() for td in self.modify],
            "delete": [td.to_dict() for td in self.delete],
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict.__contains__("changeCount") \
                and json_dict["type"] == "TDChangeset":
            changeset = Changeset(json_dict["id"], json_dict["changeCount"])
            if json_dict.__contains__("datasetId"):
                changeset.dataset_id = json_dict["datasetId"]
            if json_dict.__contains__("version"):
                changeset.version = json_dict["version"]
            if json_dict.__contains__("createdTime"):
                changeset.created_time = json_dict["createdTime"]
            if json_dict.__contains__("add"):
                changeset.add = [TrainingData.from_dict(td) for td in json_dict["add"]]
            if json_dict.__contains__("modify"):
                changeset.add = [TrainingData.from_dict(td) for td in json_dict["modify"]]
            if json_dict.__contains__("delete"):
                changeset.add = [TrainingData.from_dict(td) for td in json_dict["delete"]]
            return changeset
        else:
            raise ValueError("The given json_dict is not a valid Changeset")


@dataclass
class TrainingDataset:
    """
    Basic training dataset type
    """
    id: str
    name: str
    description: str
    license: str
    doi: str = field(default=None)
    scope: Scope = field(default=None)
    version: str = field(default=None)
    amount_of_training_data: int = field(default=None)
    created_time: str = field(default=None)
    updated_time: str = field(default=None)
    providers: List[str] = field(default_factory=list)
    keywords: List[str] = field(default_factory=list)
    metrics_in_literature: List[MetricsInLiterature] = field(default_factory=list[MetricsInLiterature])
    statistics_info: List[KeyValuePair] = field(default_factory=list[KeyValuePair])
    dataSources: List[str] = field(default_factory=list)
    number_of_classes: int = field(default=None)
    classification_schema: str = field(default=None)
    classes: Union[List[KeyValuePair], List[str]] = field(default=None)
    tasks: List[Task] = field(default_factory=list[Task])
    labeling: List[Labeling] = field(default_factory=list[Labeling])
    quality: DataQuality = field(default=None)
    changesets: List[Changeset] = field(default_factory=list[Changeset])
    data: List[TrainingData] = field(default_factory=list[TrainingData])

    def get_training_data(self) -> List[TrainingData]:
        """
        Return the training data of the training dataset
        """
        return self.data

    def get_training_data_by_id(self, _id: str) -> TrainingData:
        """
        Get training data of the training dataset by id
        """
        for data in self.data:
            if data.id == _id:
                return data

    def get_labelings(self) -> List[Labeling]:
        """
        Return the labeling of the training dataset
        """
        return self.labeling

    def get_tasks(self) -> List[Task]:
        """
        Return the tasks of the training dataset
        """
        return self.tasks

    def get_quality(self) -> DataQuality:
        """
        Return the quality of the training dataset
        """
        return self.quality

    def get_changesets(self) -> List[Changeset]:
        """
        Return the changesets of the training dataset
        """
        return self.changesets

    def to_dict(self):
        return {
            "type": "TrainingDataset",
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "license": self.license,
            "doi": self.doi,
            "scope": self.scope.to_dict() if self.scope is not None else None,
            "version": self.version,
            "amountOfTrainingData": self.amount_of_training_data,
            "createdTime": self.created_time,
            "updatedTime": self.updated_time,
            "providers": self.providers,
            "keywords": self.keywords,
            "metricsInLIT": [m.to_dict() for m in self.metrics_in_literature],
            "statisticsInfo": [s.to_dict() for s in self.statistics_info],
            "numberOfClasses": self.number_of_classes,
            "classificationSchema": self.classification_schema,
            "classes": self.classes,
            "tasks": [t.to_dict() for t in self.tasks],
            "labeling": [ll.to_dict() for ll in self.labeling],
            "quality": self.quality.to_dict() if self.quality is not None else None,
            "changesets": [changeset.to_dict() for changeset in self.changesets],
            "data": [d.to_dict() for d in self.data]
        }

    @staticmethod
    def from_dict(json_dict: dict):
        if json_dict.__contains__("id") and json_dict.__contains__("name") \
                and json_dict.__contains__("description") and json_dict.__contains__("license") \
                and json_dict["type"] == "TrainingDataset":
            td = TrainingDataset(json_dict["id"], json_dict["name"], json_dict["description"], json_dict["license"])
            if json_dict.__contains__("doi"):
                td.doi = json_dict["doi"]
            if json_dict.__contains__("scope"):
                td.scope = Scope.from_dict(json_dict["scope"])
            if json_dict.__contains__("version"):
                td.version = json_dict["version"]
            if json_dict.__contains__("amountOfTrainingData"):
                td.amount_of_training_data = json_dict["amountOfTrainingData"]
            if json_dict.__contains__("createdTime"):
                td.created_time = json_dict["createdTime"]
            if json_dict.__contains__("updatedTime"):
                td.updated_time = json_dict["updatedTime"]
            if json_dict.__contains__("providers"):
                td.providers = json_dict["providers"]
            if json_dict.__contains__("keywords"):
                td.keywords = json_dict["keywords"]
            if json_dict.__contains__("metricsInLIT"):
                td.metrics_in_literature = [MetricsInLiterature.from_dict(m) for m in
                                            json_dict["metricsInLIT"]]
            if json_dict.__contains__("statisticsInfo"):
                td.statistics_info = [KeyValuePair.from_dict(s) for s in json_dict["statisticsInfo"]]
            if json_dict.__contains__("numberOfClasses"):
                td.number_of_classes = json_dict["numberOfClasses"]
            if json_dict.__contains__("classificationSchema"):
                td.classification_schema = json_dict["classificationSchema"]
            if json_dict.__contains__("classes"):
                td.classes = json_dict["classes"]
            if json_dict.__contains__("tasks"):
                td.tasks = [Task.from_dict(t) for t in json_dict["tasks"]]
            if json_dict.__contains__("labeling"):
                td.labeling = [Labeling.from_dict(labeling) for labeling in json_dict["labeling"]]
            if json_dict.__contains__("quality"):
                td.quality = DataQuality.from_dict(json_dict["quality"])
            if json_dict.__contains__("changesets"):
                td.changesets = [Changeset.from_dict(changeset) for changeset in json_dict["changesets"]]
            if json_dict.__contains__("data"):
                td.data = [TrainingData.from_dict(data) for data in json_dict["data"]]
            return td
        else:
            raise ValueError("The given json_dict is not a valid TrainingDataset")



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------

"""Install pytdml."""

from setuptools import find_packages, setup
import io
import os.path


def read(filename, encoding='utf-8'):
    """read file contents"""
    full_path = os.path.join(os.path.dirname(__file__), filename)
    with io.open(full_path, encoding=encoding) as fh:
        contents = fh.read().strip()
    return contents


# get version number
with open(os.path.join(os.path.dirname(__file__), 'pytdml/__init__.py')) as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break

# use README.md for project long_description
long_description = read('README.md')

setup(
    name='pytdml',
    version=version,
    description='Parsing and encoding training datasets based on OGC Training Data Markup Language for AI ('
                'TrainingDML-AI) standard',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Boyi Shangguan',
    author_email='sgby@whu.edu.cn',
    url='https://github.com/TrainingDML/pytdml',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=["dataclasses", "geojson"],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    tests_require=['pytest']
)



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import cv2
import numpy as np
import tensorflow as tf
import tensorflow_io as tfio

import pytdml.utils as utils


def _parse_image(filename):
    """
    Parse a single image file.
    """
    file_content = tf.io.read_file(filename)
    if filename.endswith('.tif') or filename.endswith('.tiff'):
        return tfio.experimental.image.decode_tiff(file_content)
    elif filename.endswith('.png'):
        return tf.image.decode_png(file_content)
    elif filename.endswith('.jpg') or filename.endswith('.jpeg'):
        return tf.image.decode_jpeg(file_content)
    elif filename.endswith('.gif'):
        return tf.image.decode_gif(file_content)
    elif filename.endswith('.bmp'):
        return tf.image.decode_bmp(file_content)
    else:
        raise ValueError('Unsupported image format: {}'.format(filename))


def _parse_function_scene(file_image, label):
    return _parse_image(file_image), label


def _parse_function_object(file_image, targets):
    img = _parse_image(file_image)
    img_height = img.shape[0]
    img_width = img.shape[1]
    for target in targets:
        target[0] = target[0] / img_width
        target[1] = target[1] / img_height
        target[2] = target[2] / img_width
        target[3] = target[3] / img_height
    return img, targets


def _parse_function_segmentation(file_image, file_label_image, color_to_index):
    img = _parse_image(file_image)
    label = cv2.imread(file_label_image, cv2.COLOR_BGR2RGB)
    index_label = utils.label_to_index(label, color_to_index)
    index_label = np.asarray(index_label, dtype=np.uint8)
    index_label = tf.convert_to_tensor(index_label)
    return img, index_label


class TensorflowEOImageSceneTD:
    """
   TensorFlow Dataset for EO image scene classification training dataset
   """

    def __init__(self, td_list, class_map, resize=28):
        self.td_list = td_list
        self.class_map = class_map

    def __len__(self):
        return len(self.td_list)

    def create_dataset(self):
        """
        Create a tensorflow dataset
        """
        img_list = []
        label_list = []
        for td in self.td_list:
            img_list.append(td.data_url)
            label_list.append(self.class_map[td.labels[0].label_class])
        tf_img_list = tf.constant(img_list)
        tf_label_list = tf.constant(label_list)
        dataset = tf.data.Dataset.from_tensor_slices((tf_img_list, tf_label_list))
        dataset = dataset.map(_parse_function_scene)
        return dataset


class TensorflowEOImageObjectTD:
    """
   TensorFlow Dataset for EO image object detection training dataset
   """

    def __init__(self, td_list, class_map):
        self.td_list = td_list
        self.class_map = class_map

    def __len__(self):
        return len(self.td_list)

    def create_dataset(self):
        """
        Create a tensorflow dataset
        """
        img_list = []
        targets_list = []
        for td in self.td_list:
            img_list.append(td.data_url)
            target_list = []
            for label in self.td_list.labels:
                class_value_ = self.class_map[label.label_class]
                target_ = utils.get_bounding_box(label.object)
                target_.append(class_value_)
                target_list.append(target_)
        tf_img_list = tf.constant(img_list)
        tf_targets_list = tf.constant(targets_list)
        dataset = tf.data.Dataset.from_tensor_slices((tf_img_list, tf_targets_list))
        dataset = dataset.map(_parse_function_object)
        return dataset


class TensorflowEOImageSegmentationTD:
    """
   TensorFlow Dataset for EO image semantic segmentation training dataset
   """

    def __init__(self, td_list, class_map):
        self.td_list = td_list
        self.color_to_index = utils.class_to_index(class_map)

    def __len__(self):
        return len(self.td_list)

    def create_dataset(self):
        """
        Create a tensorflow dataset
        """
        img_list = []
        label_img_list = []
        for td in self.td_list:
            img_list.append(td.data_url)
            label_img_list.append(td.labels[0].image_url)
        tf_img_list = tf.constant(img_list)
        tf_label_img_list = tf.constant(label_img_list)
        tf_color_to_index = tf.constant(self.color_to_index)
        dataset = tf.data.Dataset.from_tensor_slices((tf_img_list, tf_label_img_list, tf_color_to_index))
        dataset = dataset.map(_parse_function_segmentation)
        return dataset



# ------------------------------------------------------------------------------
#
# Project: pytdml
# Authors: Boyi Shangguan, Kaixuan Wang
# Created: 2022-05-04
# Email: sgby@whu.edu.cn
#
# ------------------------------------------------------------------------------
#
# Copyright (c) 2022 OGC Training Data Markup Language for AI Standard Working Group
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ------------------------------------------------------------------------------
import re

import numpy as np
import geojson
from typing import Iterable


def json_empty(item):
    """
    Check if a json item is empty
    Only null, empty collections and empty strings are considered "json empty"
    """
    if item is None:
        return True
    elif isinstance(item, Iterable):
        return not item
    elif isinstance(item, str):
        return item == ""
    else:
        return False


def remove_empty(item):
    """
    Remove empty items from a json item
    """
    if isinstance(item, dict):
        new_item = {k: remove_empty(v) for k, v in item.items()}
        return {k: v for k, v in new_item.items() if not json_empty(v)}
    elif isinstance(item, (list, tuple)):
        new_item = [remove_empty(v) for v in item]
        return [v for v in new_item if not json_empty(v)]
    else:
        return item


def get_bounding_box(geometry):
    """
    Get bbox from geojson geometry object
    """
    coords = np.array(list(geojson.utils.coords(geometry)))
    return [coords[:, 0].min(), coords[:, 1].min(), coords[:, 0].max(), coords[:, 1].max()]


def class_to_index(class_map: dict):
    index = 1
    color_to_index = {}
    if len(class_map) == 2:
        for class_ in class_map:
            findall = re.findall(r'[(](.*?)[)]', class_map[class_])
            color_to_index[eval('(' + findall[0] + ')')[0]] = index
            index = index + 1
    else:
        for class_ in class_map:
            findall = re.findall(r'[(](.*?)[)]', class_map[class_])
            color_to_index[eval('(' + findall[0] + ')')] = index
            index = index + 1
    return color_to_index


def label_to_index(label, color_to_index):
    if len(label.shape) == 3:
        return label_to_index_image(label, color_to_index)
    elif len(label.shape) == 2:
        return gray_to_index_image(label, color_to_index)


def label_to_index_image(label, color_to_index):
    """
    transform 3 channel RGB label image to 1 channel index label image
    """
    assert len(label.shape) == 3
    height, width, channel = label.shape
    index_mat = np.zeros((height, width), dtype=np.uint)
    for rgb_value, class_id in color_to_index.items():
        r_, g_, b_ = rgb_value
        mask_int = (label[:, :, 0] == r_).astype(int) + \
                   (label[:, :, 1] == g_).astype(int) + \
                   (label[:, :, 2] == b_).astype(
                       int)
        mask = mask_int == 3
        index_mat[mask] = class_id
    return index_mat


def gray_to_index_image(label, color_to_index):
    """
    transform 2 channel gray value label image to 1 channel index label image
    """
    assert len(label.shape) == 2
    height, width = label.shape
    index_mat = np.zeros((height, width), dtype=np.uint)
    for gray_value, class_id in color_to_index.items():
        mask_int = label[:, :] == gray_value
        index_mat[mask_int] = class_id
    return index_mat
